--[[
Interface: 4.0.0
Title: LibBalancePowerTracker
Version: 1.0.4
Author: Kurohoshi (EU-Minahonda)

--INFO
	LibBalancePowerTracker is a library designed to provide the foresee energy feature to Balance Druids.
	CPU is only used when required, so the CPU usage is very low.
	
	FORESEE ENERGY:
	Foresee Energy is a feature that analizes the spells you have cast and/or you are casting but are yet to land and computes 
	the energy sum of them. This allows the library to distinguish between two kinds of energy and Eclipse direction: One real, 
	the one you have at the moment and other virtual, the one you'll have when all flying spells and the spell you are casting land.
	Foresee Energy works assuming the following: 
		-You're hit capped (All your spells will land).
		-You're not going to proc Euphoria (2x energy gain).
			If you proc it, it will update immediately, this only means you will reach Eclipse earlier than the library predicted the moment 
			before the Eupforia proc. 
	All the features with the 'virtual' tag (virtual Energy, virtual Eclipse ...) rely on Foresee Energy.
	
--API
	There are 5 variables related to Eclipse energy used in this addon:
		energy: The energy you have at the moment. Int = [-100,100]
		direction: The direction of the arrow. String = {"none","sun","moon"}
		virtual_energy: The energy you will have when the spell you're casting and all the flying spells land. Int = [-100,100]
		virtual_direction: The direction of the arrow when the spell you're casting and all the flying spells land. String = {"none","sun","moon"}
		reach_end: If you reach 100 or -100 virtual energy when calculating the virtual_energy, the value is stored here, false otherwise.

	FUNCTIONS:

	id = LibBalancePowerTracker:RegisterFullCallback(function(energy, direction, virtual_energy, virtual_direction, reach_end))
	These callbacks will be fired when there is a change in one of the Eclipse energy variables, usually twice per WR/SF/SS.
	NOTE: When registering a callback, that callback will be fired once.
		
	id = LibBalancePowerTracker:RegisterReducedCallback(function(energy, direction))
	These callbacks will be fired when there is a change ONLY in energy or direction, usually once per WR/SF/SS.
	NOTE: When registering a callback, that callback will be fired once.
	
	id = LibBalancePowerTracker:RegisterEclipseProbCallback(function(value))
	These callbacks will be fired every time value changes, value is the probability of Eclipse, taking into consideration euphoria and miss chance, 
	negative for Lunar Eclipse, positive for Solar Eclipse.
	NOTE: When registering a callback, that callback will be fired once.
	
	id = LibBalancePowerTracker:RegisterStatCallback(function(EnergyFunction))
	These callbacks will be fired every time EnergyFunction changes:
		energy, direction, virtual_energy, virtual_direction, reach_end = EnergyFunction(value) 
		Note: In this case, virtual_energy, virtual_direction and reach_end are special, EnergyFunction means:
		 "You have <value*100>% chance of having at least <select(3,EnergyFunction(value))> energy when all spells land, considering euphoria & miss chance." (the same goes the direction and reach_end)
	
	NOTE: When registering a callback, that callback will be fired once.
	
	failed = LibBalancePowerTracker:UnregisterCallback(id)
	Tries to unregister the callback with identifier id (id is returned only when you register the callback).
	
	energy, direction, virtual_energy, virtual_direction, reach_end = LibBalancePowerTracker:GetEclipseEnergyInfo()
	Gets the current state of the variables.
	
	value = LibBalancePowerTracker:GetEclipseChance()
	Gets the current Eclipse chance, checking Euphoria and miss chance of spells.
	
	EnergyFunction = LibBalancePowerTracker:GetEnergyFunction()	
		energy, direction, virtual_energy, virtual_direction, reach_end = EnergyFunction(value) 
		Note: In this case, virtual_energy, virtual_direction and reach_end are special, EnergyFunction means:
		 "You have <value*100>% chance of having at least <select(3,EnergyFunction(value))> energy when all spells land, considering euphoria & miss chance." (the same goes the direction and reach_end)
		
	version,subversion,revision = LibBalancePowerTracker:GetVersion()
	Gets the current working version of the library.	

--CHANGELOG
v 1.0.4 Changed Euphoria chance based on Hamlet's findings on www.elitistjerks.com 
		Added target of target to the unit check.
		Fixed casting glyph of SS counting as casting SS
		Use spellId instead of names
		
v 1.0.3 Changed to use propperly SpellQueueADT 1.1.2
		Now erases flying spells when teleporting.
		FEATURE: Eclipse chance calculation.
		FEATURE: Energy statistically calculation.

v 1.0.2 Fixed sometimes not fetching the direction properly. 
		Fixed PvP bonus

v 1.0.1 Reduced the number of callbacks fired.

v 1.0.0 Release
--]]

local version = {1,0,4};

if (LibBalancePowerTracker and LibBalancePowerTracker:CompareVersion(version)) then return; end;

--Initialize Global Lib
LibBalancePowerTracker = {};

function LibBalancePowerTracker:CompareVersion(versionTable) 
	for i,v in ipairs(versionTable) do
		if version[i] < v then
			return false;
		end;
	end;
	return true;
end;

--Locals
local LBPT = {};
local GetEclipseDirection,UnitPower,playerGUID,timeShown,callbacks,lastCallback,reducedCallbacks,elements,eclipseProbCallbacks,statCallbacks = GetEclipseDirection,UnitPower,false,GetTime(),{},0,{},0,{},{};
local options = {
	enabled = true,
	foresee = true,
}
local callBacksActivated = {
	reduced = false,
	full = false,
	eclipseProb = false,
	stat = false,
}
local vars = {
	isDruid = select(2,UnitClass('player'))=="DRUID",
	isBalance = false,
	changedState = false,
	computedEnergy=0,
	computedVirtualEnergy = 0,
	computedReachEnd = false,
	direction = "none",
	vDirection = "none",
	spellQ = SpellQueueADT:New(),
	
	unitLevelSent = 0,
	unitIsPCSent = true,
	euphoria = 0.12,
	wayTable = {},
	max_levels = 5, --Nmero mximo de hechizos que se calculan, consumo mximo aumenta de forma exponencial, 3^5 me parece un buen lmite
	
	inverseCumulativeDistributionFunction = function() return 0,"none",0,"none",false; end,
	eclipseProb= 0,
}
local missTablePvE = {
	[-4] = 1,
	[-3] = .99,
	[-2] = .98,
	[-1] = .97,
	[0] = .96,
	[1] = .95,
	[2] = .94,
	[3]	= .83,
	[4]	= .72,
	[5]	= .61,
};
local missTablePvP = {
	[-4]= 1,
	[-3]= .99,
	[-2]= .98,
	[-1]= .97,
	[0]	= .96,
	[1]	= .95,
	[2]	= .94,
	[3]	= .87,
	[4]	= .80,
	[5]	= .73,
};
local data ={
	WR  = {name = GetSpellInfo(5176) ,energy = 13,spellId=5176 }, -- name & energy Wrath
	SF  = {name = GetSpellInfo(2912) ,energy = 20,spellId=2912 }, -- name & energy Starfire
	SS  = {name = GetSpellInfo(78674),energy = 15,spellId=78674}, -- name StarSurge
	EE  = {spellId = 89265}, -- Eclipse Energy spell
	SSE = {spellId = 86605}, --Starsurge Energy spell
	SuddenEclipse = {spellId = 95746}, --PvP energy proc
};
local spellsUsed = {
	[data.WR.name] = tonumber(data.WR.spellId),
	[data.SS.name] = tonumber(data.SS.spellId),
	[data.SF.name] = tonumber(data.SF.spellId),
	[data.WR.spellId] = tostring(data.WR.name),
	[data.SS.spellId] = tostring(data.SS.name),
	[data.SF.spellId] = tostring(data.SF.name),
}
local unfilteredCombatLogEnergize = {
	[data.EE.spellId]	= function(amountEnergy)
							if amountEnergy > 0 then 
								vars.spellQ:RemoveFlyingSpell(data.SF.spellId)
							else
								vars.spellQ:RemoveFlyingSpell(data.WR.spellId)
								vars.spellQ:RecalcS(amountEnergy)
							end
							LBPT:RemoveFirst0EnergySpellFlyingQueue()
							LBPT:ChangedState(true)
						end,
	[data.SSE.spellId] 	= function() 
							vars.spellQ:RemoveFlyingSpell(data.SS.spellId)
							LBPT:RemoveFirst0EnergySpellFlyingQueue()
							LBPT:ChangedState(true)
						end,
	[data.SuddenEclipse.spellId] = function() LBPT:ChangedState(true) end,
}	
local notEnergy = {
	[data.SF.spellId] = "moon",
	[data.WR.spellId] = "sun"
}
local unfilteredCombatLogTable = {
	SPELL_ENERGIZE 	= function(id,amount) id = unfilteredCombatLogEnergize[id] if id then id(amount); end; end,
	SPELL_MISSED 	= function(id)	if spellsUsed[id] then
										vars.spellQ:MissFlyingSpell(id);
										LBPT:RemoveFirst0EnergySpellFlyingQueue();
										LBPT:ChangedState();
									end;
					end,
	SPELL_DAMAGE 	= function(id)	if notEnergy[id] == GetEclipseDirection() then
										vars.spellQ:RemoveFlyingSpell(id)
										LBPT:RemoveFirst0EnergySpellFlyingQueue()
										LBPT:ChangedState()
									end;
					end,
}
local energyFromSpell={
	[data.SF.spellId]={
		moon = 0,
		sun = data.SF.energy,
		none = data.SF.energy,
	},
	[data.WR.spellId]={
		moon = function(specialIsZero) if specialIsZero then return -data.WR.energy-1 else return -data.WR.energy; end end,
		sun = 0,
		none = function(specialIsZero) if specialIsZero then return -data.WR.energy-1 else return -data.WR.energy; end end,
	},
	[data.SS.spellId]={
		moon = -data.SS.energy,
		sun = data.SS.energy,
		none = function(_,negativeEnergy) if negativeEnergy then return (-data.SS.energy); else return data.SS.energy; end end,
	},
}
local doubleEnergyFromSpell={
	[data.SF.spellId]={
		moon = 0,
		sun = 2*data.SF.energy,
		none = 2*data.SF.energy,
	},
	[data.WR.spellId]={
		moon = function(specialIsZero) if specialIsZero then return -2*data.WR.energy-1 else return -2*data.WR.energy; end end,
		sun = 0,
		none = function(specialIsZero) if specialIsZero then return -2*data.WR.energy-1 else return -2*data.WR.energy; end end,
	},
	[data.SS.spellId]={
		moon = -data.SS.energy,
		sun = data.SS.energy,
		none = function(_,negativeEnergy) if negativeEnergy then return (-data.SS.energy); else return data.SS.energy; end end,
	},
}

--Aux functions
local function UnitByName(name)
	if select(1,UnitName("target"))==name then return "target";
	elseif select(1,UnitName("mouseover"))==name then return "mouseover";
	elseif select(1,UnitName("focus"))==name then return "focus";
	elseif select(1,UnitName("player"))==name then return "player";
	elseif select(1,UnitName("targettarget"))==name then return "targettarget";
	end
	
	--boss
	for i=1,4 do if select(1,UnitName("boss"..i))==name then return "boss"..i; end end
	
	--arena
	if select(1,IsActiveBattlefieldArena()) then
		for i=1,5 do if select(1,UnitName("arena"..i))==name then return "arena"..i; end end
	end

	if GetNumRaidMembers()==0 then
		--party
		if GetNumPartyMembers() ~=0 then for i=1,5 do if select(1,UnitName("party"..i))==name then return "party"..i; end end end
	else
		--raid
		for i=1,40 do if select(1,UnitName("raid"..i))==name then return "raid"..i; end	end
	end

	return "target";
end
local function UpdateSpellcastSentEvent()
	if (next(eclipseProbCallbacks) ~= nil or next(statCallbacks) ~= nil) and options.enabled and options.foresee and vars.isBalance then
		LibBalancePowerTrackerEventFrame:RegisterEvent("UNIT_SPELLCAST_SENT");
		function LBPT.MissChance()
			local diff = vars.unitLevelSent - UnitLevel("player");
			if diff <= -4 then return 0;elseif diff > 5 then diff = 5;end
			if vars.unitIsPCSent then return 1-min(missTablePvP[diff]+GetCombatRatingBonus(8)*.01,1);end
			return 1-min(missTablePvE[diff]+GetCombatRatingBonus(8)*.01,1);
		end
	else
		LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_SENT");
		function LBPT.MissChance() return 0; end;
	end
end
local function UpdateFunctions()
	callBacksActivated.reduced,callBacksActivated.full = next(reducedCallbacks) ~= nil,next(callbacks) ~= nil;
	callBacksActivated.eclipseProb,callBacksActivated.stat = next(eclipseProbCallbacks) ~= nil,next(statCallbacks) ~= nil;
	if elements ~= 0 then
		if (callBacksActivated.eclipseProb or callBacksActivated.stat) and callBacksActivated.reduced then
			function LBPT:ChangedState(real) LBPT:RecalcEnergy(); LBPT:RecalcWays();	LBPT:FireCallbacks(); if real then LBPT:FireReducedCallbacks() end;	end;--aded reduced callbacks and stat analysis
		elseif callBacksActivated.reduced then
			function LBPT:ChangedState(real) LBPT:RecalcEnergy();						LBPT:FireCallbacks(); if real then LBPT:FireReducedCallbacks() end;	end;--aded reduced callbacks
		elseif (callBacksActivated.eclipseProb or callBacksActivated.stat) then
			function LBPT:ChangedState() 	 LBPT:RecalcEnergy(); LBPT:RecalcWays();	LBPT:FireCallbacks();												end;--added stat analysis 
		else
			function LBPT:ChangedState() 	 LBPT:RecalcEnergy();						LBPT:FireCallbacks();												end;--only full callbacks (most common)
		end
		
	else 	
			function LBPT:ChangedState() 	vars.changedState=true	end;--Default, no callbacks, just waiting till someone calls GetEclipseEnergyInfo()
	end
	UpdateSpellcastSentEvent()
end
local function UpdateEuphoria()	vars.euphoria = select(5,GetTalentInfo(1,7)) * .12; end --Checked on shapeshift_forms_updated on load, on Player_login on reload
local function EnergyFromSpell(id,special,direction,energy) --returns energy from spellid, direction, energy & special
	local tmp = energyFromSpell[id][direction]
	if tonumber(tmp) then return tmp;
	else return tmp(special==0,energy < 0); end	
end
local function TreeNormalEnergyFromSpell(id,sWR,direction,energy,rE)
	energy = energy + EnergyFromSpell(id,sWR,direction,energy);
	
	if id==data.WR.spellId then sWR = (sWR+1) % 3; end
	if energy>=100 then			return    100,   "moon",  100,sWR;
	elseif energy<=-100 then 	return   -100,    "sun", -100,sWR;
	else						return energy,direction,   rE,sWR;
	end
end
local function TreeDoubleEnergyFromSpell(id,sWR,direction,energy,rE)
	local tmp = doubleEnergyFromSpell[id][direction]
	if not tonumber(tmp) then tmp = tmp(special==0,energy < 0); end	
	
	energy = energy + tmp;
	if id==data.WR.spellId then sWR = (sWR+2) % 3; end
	if energy>=100 then			return    100,   "moon",  100,sWR;
	elseif energy<=-100 then 	return   -100,    "sun", -100,sWR;
	else						return energy,direction,   rE,sWR;
	end
end
local function TreeApplySpell(parentOriginal,sonMiss,sonNormal,sonDouble,element)
	local prob, energy, direction, reachEnd, sWR = parentOriginal.prob,parentOriginal.energy,parentOriginal.direction,parentOriginal.reachEnd, parentOriginal.sWR;
		
	if prob == 0 then sonMiss.prob,sonNormal.prob,sonDouble.prob = 0,0,0; return; end
	
	local euphoriaProb = (((abs(energy)<41) and (not ((direction=="moon" and energy>0) or (direction=="sun" and energy<0)))) and vars.euphoria) or 0;
	sonMiss.prob = prob * element.mc;
	sonDouble.prob = prob *(1-element.mc)* euphoriaProb 
	sonNormal.prob = prob *(1-element.mc)*(1-euphoriaProb)
	
	if sonMiss.prob ~=0 then sonMiss.energy,sonMiss.direction,sonMiss.reachEnd,sonMiss.sWR = energy,direction,reachEnd,sWR;end
	if sonDouble.prob~=0 then sonDouble.energy,sonDouble.direction,sonDouble.reachEnd,sonDouble.sWR = TreeDoubleEnergyFromSpell(element.n,sWR,direction,energy,reachEnd);end
	if sonNormal.prob~=0 then sonNormal.energy,sonNormal.direction,sonNormal.reachEnd,sonNormal.sWR = TreeNormalEnergyFromSpell(element.n,sWR,direction,energy,reachEnd);end
end
function LBPT.MissChance() return 0;end --will be modified

--Loading
local frame = CreateFrame("Frame","LibBalancePowerTrackerEventFrame",UIParent);
frame:SetScript("OnEvent",  function(self, event, ...) 	LBPT[event](self,...)	end);
frame:SetScript("OnUpdate", function() if UnitPower("player",8) ~= 0 then frame:Hide(); LBPT:ChangedState(true); elseif (timeShown < GetTime()) or (elements == 0) then frame:Hide(); end; end); 
frame:SetScript("OnShow",	function() timeShown = GetTime()+1.5; end); 
frame:RegisterEvent("PLAYER_LOGIN"); 
function LBPT:PLAYER_LOGIN() playerGUID=UnitGUID("player"); UpdateEuphoria() LBPT:ReCheck(); end;

function LBPT:ReCheck() 
	vars.isBalance = GetSpellCooldown(data.SS.name)~=nil;

	if options.enabled and vars.isDruid then
		LibBalancePowerTrackerEventFrame:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");	 
		LibBalancePowerTrackerEventFrame:RegisterEvent("CHARACTER_POINTS_CHANGED");	
		LibBalancePowerTrackerEventFrame:RegisterEvent("UPDATE_SHAPESHIFT_FORMS");
	else
		LibBalancePowerTrackerEventFrame:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED");	 
		LibBalancePowerTrackerEventFrame:UnregisterEvent("CHARACTER_POINTS_CHANGED");
		LibBalancePowerTrackerEventFrame:UnregisterEvent("UPDATE_SHAPESHIFT_FORMS");		
	end
		
	if options.enabled and vars.isBalance then
		if options.foresee then
			LibBalancePowerTrackerEventFrame:RegisterEvent("UNIT_SPELLCAST_START");
			LibBalancePowerTrackerEventFrame:RegisterEvent("UNIT_SPELLCAST_FAILED");
			LibBalancePowerTrackerEventFrame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED");
			LibBalancePowerTrackerEventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
			function LBPT:COMBAT_LOG_EVENT_UNFILTERED( _,event,gUIDor,_,_,_,_,_,spellId,_,_,amountEnergy)
				if (gUIDor ~= playerGUID) then return end event = unfilteredCombatLogTable[event] if event then event(spellId,amountEnergy) end
			end
		else
			LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_START");
			LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_FAILED");
			LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED");
			LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED");
			function LBPT:COMBAT_LOG_EVENT_UNFILTERED( _,event,gUIDor,_,_,_,_,_,_,_,_,_,typeEnergy)
				if (gUIDor == playerGUID) and (event=="SPELL_ENERGIZE") and (typeEnergy == 8) then LBPT:ChangedState(true) end;
			end
		end
		LibBalancePowerTrackerEventFrame:RegisterEvent("PLAYER_DEAD");
		
		LibBalancePowerTrackerEventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");	
		LibBalancePowerTrackerEventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	else 
		LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_START");
		LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_FAILED");
		LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED");
		LibBalancePowerTrackerEventFrame:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED");
		LibBalancePowerTrackerEventFrame:UnregisterEvent("PLAYER_DEAD");
		
		LibBalancePowerTrackerEventFrame:UnregisterEvent("PLAYER_ENTERING_WORLD");	
		LibBalancePowerTrackerEventFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
		
		vars.computedEnergy,vars.direction,vars.computedVirtualEnergy,vars.vDirection,vars.computedReachEnd = 0,"none",0,"none",false;
		vars.inverseCumulativeDistributionFunction = function() return 0,"none",0,"none",false; end
		vars.eclipseProb = 0
	end
	
	UpdateSpellcastSentEvent()
	
	vars.spellQ:Clear();
	LBPT:ChangedState(true);
end

--Combat events
function LBPT:UNIT_SPELLCAST_SENT(unitID,name,_,target) 
	if unitID == "player" and spellsUsed[name] then
		unitID = UnitByName(target)
		vars.unitIsPCSent	= UnitIsPlayer(unitID);
		vars.unitLevelSent	= UnitLevel(unitID);
		if vars.unitLevelSent == -1 then vars.unitLevelSent = UnitLevel("player") + 3; end
	end
end
function LBPT:UNIT_SPELLCAST_START(unit,_,_,num,id)
	if unit == "player" and spellsUsed[id] then
		vars.spellQ:BeginCastingSpell(id,num,LBPT.MissChance())
		LBPT:ChangedState()
	end
end
function LBPT:UNIT_SPELLCAST_SUCCEEDED(unit,_,_,num,id)
	if unit == "player" and spellsUsed[id] and vars.spellQ:FinishCastingSpell(id,num,LBPT.MissChance()) then
		LBPT:ChangedState()
	end
end
function LBPT:UNIT_SPELLCAST_FAILED(unit,_,_,num)		if unit == "player" and vars.spellQ:FailedCastingSpell(num) then LBPT:ChangedState() end end
function LBPT:UNIT_SPELLCAST_INTERRUPTED(unit,_,_,num) 	if unit == "player" and vars.spellQ:FailedCastingSpell(num) then LBPT:ChangedState() end end

function LBPT:COMBAT_LOG_EVENT_UNFILTERED( _,event,gUIDor,_,_,_,_,_,spellId,_,_,amountEnergy) 
	if (gUIDor ~= playerGUID) then return end
	event = unfilteredCombatLogTable[event]
	if event then event(spellId,amountEnergy) end
end

function LBPT:PLAYER_DEAD() LibBalancePowerTrackerEventFrame:RegisterEvent("PLAYER_ALIVE"); vars.spellQ:Clear()	LBPT:ChangedState(true);	end; --Reset queue & clear energy when die

--Issues with dying because of a crit with 4p PvP bonus 
function LBPT:PLAYER_ALIVE() 
	LibBalancePowerTrackerEventFrame:UnregisterEvent("PLAYER_ALIVE"); 
	if (UnitPower("player",8) ~= 0) then	
		if not vars.pvpFrame then
			vars.pvpFrame = CreateFrame("Frame","LibBalancePowerTrackerPvPFrame",UIParent);
			vars.pvpFrame:SetScript("OnUpdate", function() if UnitPower("player",8)==0 then vars.pvpFrame:Hide(); LBPT:ChangedState(true);end;end); 
		end
		vars.pvpFrame:Show()
	end;
end; 

--Direction & energy when teleporting 
function LBPT:PLAYER_ENTERING_WORLD()		LibBalancePowerTrackerEventFrame:RegisterEvent("ECLIPSE_DIRECTION_CHANGE");		vars.spellQ:Clear();	LBPT:ChangedState(true); end; 
function LBPT:ECLIPSE_DIRECTION_CHANGE()	LibBalancePowerTrackerEventFrame:UnregisterEvent("ECLIPSE_DIRECTION_CHANGE");							LBPT:ChangedState(true); end; 

--Talent change events
function LBPT:ACTIVE_TALENT_GROUP_CHANGED()	UpdateEuphoria(); LBPT:ReCheck(); end; 
function LBPT:CHARACTER_POINTS_CHANGED() 	UpdateEuphoria(); LBPT:ReCheck(); end;
--Talent check events
function LBPT:UPDATE_SHAPESHIFT_FORMS()		UpdateEuphoria(); end;

--Spell queue TAD
local function RemoveFirst0EnergySpellFlyingQueueRecursive(direction,energy)
	local id,special = vars.spellQ:First()
	if id and EnergyFromSpell(id,special,direction,energy)==0 then
		vars.spellQ:RemoveFlyingSpell(id)
		RemoveFirst0EnergySpellFlyingQueueRecursive(direction,energy)
	end
end
function LBPT:RemoveFirst0EnergySpellFlyingQueue()
	RemoveFirst0EnergySpellFlyingQueueRecursive(GetEclipseDirection(),UnitPower("player" , 8))
end

--ChangedState
function LBPT:ChangedState() vars.changedState=true end;--will be changed when registeriong callbacks (UpdateFunctions())

--Recalc Energy function
local function ExtraEnergy(energy,direction) --computes extra energy
	local reachEnd,getNext = false,vars.spellQ:iterator();
	local v = getNext();
	
	while v do
		energy = energy + EnergyFromSpell(v.n,v.s,direction,energy)
		v = getNext();
	
		if energy>=100 then 
			energy = 100
			direction = "moon"
			reachEnd = energy
		elseif energy<=-100 then 
			energy = -100 
			direction = "sun";
			reachEnd = energy;
		end
	end

	return energy,direction,reachEnd
end
function LBPT:RecalcEnergy()
	local direction,energy = GetEclipseDirection(),UnitPower("player" , 8);
	if energy == 0 then LibBalancePowerTrackerEventFrame:Show(); end
	
	vars.computedEnergy,vars.direction,vars.changedState = energy, direction, false;
	
	if options.foresee then
		vars.spellQ:RemoveTimedOutFlyingSpell(6);
		vars.computedVirtualEnergy,vars.vDirection,vars.computedReachEnd = ExtraEnergy(energy,direction)
		return
	end
	
	vars.computedVirtualEnergy,vars.vDirection,vars.computedReachEnd = energy,direction,false;
end

local checkNextTable = {
	["-100"] = {	
		["-100"] = function(node,now) return node.energy < now.energy end,
		["false"] = false,
		["100"] = false,
	},
	["false"] = {
		["-100"] = true,
		["false"] = function(node,now) return node.energy > now.energy end,
		["100"] = false,
	},
	["100"] = {	
		["-100"] = true,
		["false"] = true,
		["100"] = function(node,now) return node.energy < now.energy end,
	},
}
local function checkNext(node,now) --return true si no debemos insertar, sino pasar al siguiente
	local temp = checkNextTable[tostring(node.reachEnd)][tostring(now.reachEnd)]
	if type(temp) == "boolean" then return temp else return temp(node,now) end
end
function LBPT:RecalcWays(forced)
	local direction,energy,getNext = GetEclipseDirection(),UnitPower("player" , 8),vars.spellQ:iterator();
	
	--Aclarar que tenemos en cuenta que el %miss se calcula cuando el hechizo se enva, tambin contamos con 
	--que euforia no acta como la tabla de ataques mele (se hacen dos tiradas independientes, la primera para 
	--ver si el hechizo acierta y la segunda para ver si salta euforia)
	
	--[[ WAYTREE as a Table (First child is miss, second is normal and third is Euphoria
	--(1)(2)(3)(4)(5)(6)(7)(8)(9)(10)(11)(12)(13)(14)(15)(16)(17)(18)(19)(20)(21)(22)(23)(24)(25)(26)(27)--]]
	
	local tempTable,wayEnd,v,level,idFirstSpell  = vars.wayTable, 1,getNext(),0,false;
	if not tempTable[1] then tempTable[1] = {} end
		if v then tempTable[1].sWR,idFirstSpell = v.s,v.n end
		tempTable[1].prob = 1;
		tempTable[1].energy = energy;
		tempTable[1].direction = direction;
		tempTable[1].reachEnd = false;
	
	while v and level <= vars.max_levels do
		for i = 1,wayEnd do 
			if not tempTable[i+wayEnd] then  tempTable[i+wayEnd] = {} end
			if not tempTable[i+2*wayEnd] then tempTable[i+2*wayEnd] = {} end
			TreeApplySpell(tempTable[i],tempTable[i], tempTable[i+wayEnd] , tempTable[i+2*wayEnd] ,v)
		end
		wayEnd,v,level=wayEnd*3,getNext(),level+1;
	end
	
	if callBacksActivated.eclipseProb or forced then --Eclipse probability calc if only the eclipse calc is selected
		local a,rE = 0,false
		for i=1,wayEnd do 
			if tempTable[i].prob>0 and tempTable[i].reachEnd then
				rE,a = rE or tempTable[i].reachEnd,a + tempTable[i].prob;
			end;
		end;
		a = (rE and (a * rE/100)) or 0;
		if vars.eclipseProb ~= a then --Eclipse chance has changed
			vars.eclipseProb = a;for k,v in pairs(eclipseProbCallbacks) do v(a);end;
		end
	end
	
	if callBacksActivated.stat or forced then
		--funcin de distribucin como lista doblemente enlazada (cumulative distribution as doubly linked list) 
		local first = false;
		local last;
		local now;
		
		for i = 1,wayEnd do 
			if tempTable[i].energy and tempTable[i].prob>0 then
				if not first then --si est vaca la inicializamos
					first=tempTable[i];
					last =tempTable[i];
					tempTable[i].nextItem = nil;
					tempTable[i].previousItem = nil;		
				else --si no est vaca, nos ponemos en el primero y evaluamos, si tenemos que insertar, insertamos, si no, pasamos al siguiente
					now = first;
					while now do
						if checkNext(tempTable[i],now) then  --no debemos insertar, sino pasar al siguiente
							if now.nextItem then  --hay siguiente																			
								now = now.nextItem;
							else -- no hay siguiente, insertamos al final de la lista
								tempTable[i].previousItem = now;
								tempTable[i].nextItem = nil;
								now.nextItem = tempTable[i]
								last = tempTable[i];
								now = false;
							end
						elseif (tempTable[i].energy == now.energy) and (tempTable[i].reachEnd == now.reachEnd) then --no insertamos, actualizamos el valor
							now.prob=now.prob+tempTable[i].prob
							now = false;
						else -- hay que insertar, moviendo el item una posicion a la derecha
							if now.previousItem then --hay anterior (no es el primero), hay que desplazar los valores una posicin a la derecha
								tempTable[i].previousItem = now.previousItem;
								tempTable[i].nextItem = now;
								now.previousItem.nextItem = tempTable[i];
								now.previousItem = tempTable[i];
								now=false;
							else --es el primero, hay que insertar al principio de la tabla 
								tempTable[i].previousItem = nil;
								tempTable[i].nextItem = now;
								first=tempTable[i];
								now.previousItem = tempTable[i];
								now=false;
							end
						end
					end 
				end
			end
		end

		--[[now=first;
		print("BEGIN")
		while now do
			print(now.energy..": "..tostring(now.reachEnd).." = "..now.prob)
			now = now.nextItem;
		end
		print("END")--]]
		
		--funcion (valor) devuelve la mnima energa que tendrs con un valor% de seguridad
		-- "tienes un valor*100% de tener como mnimo funcion(valor) energa
		if (direction == "sun") then --empiezo comprobando probabilidades desde el final
			vars.inverseCumulativeDistributionFunction = 	function(value) if value>1 then value = 1 elseif value <0 then value = 0 end local last = last;		while last do value=value-last.prob;	if value<=0 then return energy,direction,last.energy,last.direction,last.reachEnd;		else last = last.previousItem;	end end	end	
			for k,v in pairs(statCallbacks) do v(vars.inverseCumulativeDistributionFunction);end;
		elseif (direction == "moon") then --empiezo comprobando probabilidades desde el principio
			vars.inverseCumulativeDistributionFunction = 	function(value) if value>1 then value = 1 elseif value <0 then value = 0 end local first = first;	while first do value=value-first.prob;	if value<=0 then return energy,direction,first.energy,first.direction,first.reachEnd;	else first = first.nextItem 	end	end	end	
			for k,v in pairs(statCallbacks) do v(vars.inverseCumulativeDistributionFunction);end;
		elseif (idFirstSpell == data.SF.spellId) then 
			-- = que en direction == sun
			vars.inverseCumulativeDistributionFunction = 	function(value) if value>1 then value = 1 elseif value <0 then value = 0 end local last = last;		while last do value=value-last.prob;	if value<=0 then return energy,direction,last.energy,last.direction,last.reachEnd;		else last = last.previousItem;	end end	end	
			for k,v in pairs(statCallbacks) do v(vars.inverseCumulativeDistributionFunction);end;
		elseif (idFirstSpell == data.WR.spellId) then
			-- = que en direction == moon
			vars.inverseCumulativeDistributionFunction = 	function(value) if value>1 then value = 1 elseif value <0 then value = 0 end local first = first;	while first do value=value-first.prob;	if value<=0 then return energy,direction,first.energy,first.direction,first.reachEnd;	else first = first.nextItem 	end	end	end	
			for k,v in pairs(statCallbacks) do v(vars.inverseCumulativeDistributionFunction);end;
		elseif energy <0 then
			-- = que en direction == moon
			vars.inverseCumulativeDistributionFunction = 	function(value) if value>1 then value = 1 elseif value <0 then value = 0 end local first = first;	while first do value=value-first.prob;	if value<=0 then return energy,direction,first.energy,first.direction,first.reachEnd;	else first = first.nextItem 	end	end	end	
			for k,v in pairs(statCallbacks) do v(vars.inverseCumulativeDistributionFunction);end;
		else
			-- = que en direction == sun
			vars.inverseCumulativeDistributionFunction = 	function(value) if value>1 then value = 1 elseif value <0 then value = 0 end local last = last;		while last do value=value-last.prob;	if value<=0 then return energy,direction,last.energy,last.direction,last.reachEnd;		else last = last.previousItem;	end end	end	
			for k,v in pairs(statCallbacks) do v(vars.inverseCumulativeDistributionFunction);end;
		end
	end
end

--Calling functions
function LBPT:FireCallbacks() 			for k,v in pairs(callbacks) 		do v(vars.computedEnergy,vars.direction,vars.computedVirtualEnergy,vars.vDirection,vars.computedReachEnd); 	end; end;
function LBPT:FireReducedCallbacks() 	for k,v in pairs(reducedCallbacks) 	do v(vars.computedEnergy,vars.direction); 																	end; end;
--Called functions
function LibBalancePowerTracker:GetEclipseEnergyInfo(forced)	if vars.changedState or forced then LBPT:RecalcEnergy() end return vars.computedEnergy,vars.direction,vars.computedVirtualEnergy,vars.vDirection,vars.computedReachEnd;	end;
function LibBalancePowerTracker:GetEclipseChance(forced)		if (not callBacksActivated.eclipseProb) or vars.changedState or forced then LBPT:RecalcEnergy()	LBPT:RecalcWays(not callBacksActivated.eclipseProb)	end return vars.eclipseProb; end;
function LibBalancePowerTracker:GetEnergyFunction(forced)		if (not callBacksActivated.stat) 		or vars.changedState or forced then LBPT:RecalcEnergy()	LBPT:RecalcWays(not callBacksActivated.stat )		end return vars.inverseCumulativeDistributionFunction; end;
function LibBalancePowerTracker:RegisterReducedCallback(callback)
	lastCallback=lastCallback+1
	elements=elements+1
	reducedCallbacks[lastCallback]=callback;
	local energy,direction = LibBalancePowerTracker:GetEclipseEnergyInfo()
	callback(energy,direction)
	UpdateFunctions()
	return lastCallback
end
function LibBalancePowerTracker:RegisterFullCallback(callback)
	lastCallback=lastCallback+1
	elements=elements+1
	callbacks[lastCallback]=callback;
	callback(LibBalancePowerTracker:GetEclipseEnergyInfo())
	UpdateFunctions()
	return lastCallback
end
function LibBalancePowerTracker:RegisterEclipseProbCallback(callback)
	lastCallback=lastCallback+1
	elements=elements+1
	LBPT:RecalcWays()
	eclipseProbCallbacks[lastCallback]=callback;
	callback(LibBalancePowerTracker:GetEclipseChance())
	UpdateFunctions()
	return lastCallback
end
function LibBalancePowerTracker:RegisterStatCallback(callback)
	lastCallback=lastCallback+1
	elements=elements+1
	statCallbacks[lastCallback]=callback;
	callback(LibBalancePowerTracker:GetEnergyFunction())
	UpdateFunctions()
	return lastCallback
end
function LibBalancePowerTracker:UnregisterCallback(id)
	if reducedCallbacks[id] then
		reducedCallbacks[id]=nil;
	elseif callbacks[id] then
		callbacks[id]=nil;
	elseif eclipseProbCallbacks[id] then
		eclipseProbCallbacks[id]=nil;
	elseif statCallbacks[id] then
		statCallbacks[id]=nil;
	else
		return true;
	end
	elements=elements-1
	UpdateFunctions()
end
function LibBalancePowerTracker:GetVersion()	return version[1],version[2],version[3]; end;
--function LibBalancePowerTracker:GetEnabled()	return options.enabled; end;
--function LibBalancePowerTracker:Enable(enable)	options.enabled=enable;	LBPT:ReCheck(); end;
